
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <gtk/gtk.h>

#include "defines.h"
#include "memory.h"
#include "editor.h"
#include "prefs.h"
#include "common.h"
#include "exit.h"


/* version string */
char version_string[] = "Tile Driller v2.0";


void common_copy_data_to_view(unsigned char *in, unsigned char *out, int dx, int dy, int zoom, int mode) {

  int x, y, a, b, z3, ip, op, ap, xz, yz, dxz3;
  float a1, a2, a3, a4, bg;
  unsigned char c, cr, cg, cb;


  z3 = zoom*3;
  dxz3 = dx*z3;

  if (mode == VIEW_MODE_RGB) {
    /* render RGB */
    for (y = 0; y < dy; y++) {
      for (x = 0; x < dx; x++) {
	ip = (y*dx + x)<<2;
	op = (y*zoom*dx + x)*z3;
	for (a = 0; a < zoom; a++) {
	  ap = op + a*dxz3;
	  for (b = 0; b < zoom; b++) {
	    out[ap++] = in[ip + 0];
	    out[ap++] = in[ip + 1];
	    out[ap++] = in[ip + 2];
	  }
	}
      }
    }
  }
  else if (mode == VIEW_MODE_ALPHA) {
    /* render only the alpha channel */
    for (y = 0; y < dy; y++) {
      for (x = 0; x < dx; x++) {
	ip = (y*dx + x)<<2;
	c = in[ip + 3];
	op = (y*zoom*dx + x)*z3;
	for (a = 0; a < zoom; a++) {
	  ap = op + a*dxz3;
	  for (b = 0; b < zoom; b++) {
	    out[ap++] = c;
	    out[ap++] = c;
	    out[ap++] = c;
	  }
	}
      }
    }
  }
  else if (mode == VIEW_MODE_RGBA) {
    /* render RGBA */
    yz = 0;
    for (y = 0; y < dy; y++) {
      xz = 0;
      ip = (y*dx) << 2;
      for (x = 0; x < dx; x++) {
	a1 = ((float)in[ip + 3])/255.0f;
	a2 = 1.0f - a1;
	a3 = 128.0f*a2;
	a4 = 192.0f*a2;
	op = (yz*dx + x)*z3;
	cr = ((float)in[ip + 0])*a1;
	cg = ((float)in[ip + 1])*a1;
	cb = ((float)in[ip + 2])*a1;
	for (a = 0; a < zoom; a++) {
	  ap = op + a*dxz3;
	  for (b = 0; b < zoom; b++) {
	    /* compute the background color */
	    if ((yz+a) & 8) {
	      if ((xz+b) & 8)
		bg = a3;
	      else
		bg = a4;
	    }
	    else {
	      if ((xz+b) & 8)
		bg = a4;
	      else
		bg = a3;
	    }

	    out[ap++] = cr + bg;
	    out[ap++] = cg + bg;
	    out[ap++] = cb + bg;
	  }
	}
	xz += zoom;
	ip += 4;
      }
      yz += zoom;
    }
  }
}


void common_copy_pixel_to_view(unsigned char *in, unsigned char *out, int dx, int dy, int x, int y, int zoom, int mode) {

  int a, b, z3, ip, op, ap, xz, yz, dxz3;
  float a1, a2, a3, a4, bg;
  unsigned char c, cr, cg, cb;


  z3 = zoom*3;
  dxz3 = dx*z3;

  if (mode == VIEW_MODE_RGB) {
    /* render RGB */
    ip = (y*dx + x)<<2;
    op = (y*zoom*dx + x)*z3;
    for (a = 0; a < zoom; a++) {
      ap = op + a*dxz3;
      for (b = 0; b < zoom; b++) {
	out[ap++] = in[ip + 0];
	out[ap++] = in[ip + 1];
	out[ap++] = in[ip + 2];
      }
    }
  }
  else if (mode == VIEW_MODE_ALPHA) {
    /* render only the alpha channel */
    ip = (y*dx + x)<<2;
    c = in[ip + 3];
    op = (y*zoom*dx + x)*z3;
    for (a = 0; a < zoom; a++) {
      ap = op + a*dxz3;
      for (b = 0; b < zoom; b++) {
	out[ap++] = c;
	out[ap++] = c;
	out[ap++] = c;
      }
    }
  }
  else if (mode == VIEW_MODE_RGBA) {
    /* render RGBA */
    yz = y*zoom;
    xz = x*zoom;
    ip = (y*dx + x)<<2;
    a1 = ((float)in[ip + 3])/255.0f;
    a2 = 1.0f - a1;
    a3 = 128.0f*a2;
    a4 = 192.0f*a2;
    op = (yz*dx + x)*z3;
    cr = ((float)in[ip + 0])*a1;
    cg = ((float)in[ip + 1])*a1;
    cb = ((float)in[ip + 2])*a1;
    for (a = 0; a < zoom; a++) {
      ap = op + a*dxz3;
      for (b = 0; b < zoom; b++) {
	/* compute the background color */
	if ((yz+a) & 8) {
	  if ((xz+b) & 8)
	    bg = a3;
	  else
	    bg = a4;
	}
	else {
	  if ((xz+b) & 8)
	    bg = a4;
	  else
	    bg = a3;
	}

	out[ap++] = cr + bg;
	out[ap++] = cg + bg;
	out[ap++] = cb + bg;
      }
    }
  }
}


void common_copy_block_to_view(unsigned char *in, unsigned char *out, int dx, int dy, int zoom, int mode, int ax, int ay, int mx, int my) {

  int x, y, a, b, z3, ip, op, ap, xz, yz, dxz3;
  float a1, a2, a3, a4, bg;
  unsigned char c, cr, cg, cb;


  z3 = zoom*3;
  dxz3 = dx*z3;

  mx += ax;
  my += ay;

  if (mode == VIEW_MODE_RGB) {
    /* render RGB */
    for (y = ay; y < my; y++) {
      for (x = ax; x < mx; x++) {
	ip = (y*dx + x)<<2;
	op = (y*zoom*dx + x)*z3;
	for (a = 0; a < zoom; a++) {
	  ap = op + a*dxz3;
	  for (b = 0; b < zoom; b++) {
	    out[ap++] = in[ip + 0];
	    out[ap++] = in[ip + 1];
	    out[ap++] = in[ip + 2];
	  }
	}
      }
    }
  }
  else if (mode == VIEW_MODE_ALPHA) {
    /* render only the alpha channel */
    for (y = ay; y < my; y++) {
      for (x = ax; x < mx; x++) {
	ip = (y*dx + x)<<2;
	c = in[ip + 3];
	op = (y*zoom*dx + x)*z3;
	for (a = 0; a < zoom; a++) {
	  ap = op + a*dxz3;
	  for (b = 0; b < zoom; b++) {
	    out[ap++] = c;
	    out[ap++] = c;
	    out[ap++] = c;
	  }
	}
      }
    }
  }
  else if (mode == VIEW_MODE_RGBA) {
    /* render RGBA */
    for (y = ay; y < my; y++) {
      yz = y*zoom;
      for (x = ax; x < mx; x++) {
	xz = x*zoom;
	ip = (y*dx + x)<<2;
	a1 = ((float)in[ip + 3])/255.0f;
	a2 = 1.0f - a1;
	a3 = 128.0f*a2;
	a4 = 192.0f*a2;
	op = (yz*dx + x)*z3;
	cr = ((float)in[ip + 0])*a1;
	cg = ((float)in[ip + 1])*a1;
	cb = ((float)in[ip + 2])*a1;
	for (a = 0; a < zoom; a++) {
	  ap = op + a*dxz3;
	  for (b = 0; b < zoom; b++) {
	    /* compute the background color */
	    if ((yz+a) & 8) {
	      if ((xz+b) & 8)
		bg = a3;
	      else
		bg = a4;
	    }
	    else {
	      if ((xz+b) & 8)
		bg = a4;
	      else
		bg = a3;
	    }

	    out[ap++] = cr + bg;
	    out[ap++] = cg + bg;
	    out[ap++] = cb + bg;
	  }
	}
      }
    }
  }
}


void common_gtk_exit(GtkWidget *widget, gpointer data) {

  /* display a warning if the image has not been saved */
  if (editor_need_to_save == YES) {
    gtk_widget_show_all(exit_dialog);
    return;
  }

  /* save the prefs before quitting */
  prefs_save();

  gtk_main_quit();
}


gint common_gtk_hide_widget(GtkWidget *widget, gpointer data) {

  gtk_widget_hide(widget);

  return TRUE;
}


int common_copy_image(unsigned char *out, int ox, int oy, unsigned char *in, int ix, int iy) {

  int x, y, z, i, n, ex, ey;


  if (ox < ix)
    ex = ox;
  else
    ex = ix;

  if (oy < iy)
    ey = oy;
  else
    ey = iy;

  /* first clear the output */
  for (x = 0, i = 0; x < ox*oy; x++) {
    out[i++] = editor_pen_bg_r;
    out[i++] = editor_pen_bg_g;
    out[i++] = editor_pen_bg_b;
    out[i++] = editor_pen_bg_a;
  }

  /* then copy as much as possible */
  for (y = 0, i = 0; y < ey; y++) {
    n = i;
    z = (y*ix) << 2;
    for (x = 0; x < ex; x++) {
      out[n++] = in[z + 0];
      out[n++] = in[z + 1];
      out[n++] = in[z + 2];
      out[n++] = in[z + 3];
      z += 4;
    }
    i += ox<<2;
  }

  return SUCCEEDED;
}


int common_reformat_image(int *dx, int *dy, int *bpp, unsigned char **d) {

  unsigned char *n;
  int w, h, x, y, i, p, k;


  if (dx == NULL || dy == NULL || bpp == NULL || d == NULL)
    return FAILED;

  if (*bpp == 3) {
    /* add the alpha channel */
    n = malloc((*dx)*(*dy)*4);
    if (n == NULL) {
      fprintf(stderr, "COMMON_REFORMAT_IMAGE: Out of memory error.\n");
      return FAILED;
    }

    for (k = 0, p = 0, i = 0; i < (*dx)*(*dy); i++) {
      n[p++] = (*d)[k++];
      n[p++] = (*d)[k++];
      n[p++] = (*d)[k++];
      n[p++] = 255;
    }

    free(*d);
    *d = n;
  }

  /* resize the image, if it's not of suitable size */
  if (*dx != *dy || !(*dx == 64 || *dx == 128 || *dx == 256 || *dx == 512 || *dx == 1024 || *dx == 2048)) {
    if (*dx <= 64)
      w = 64;
    else if (*dx <= 128)
      w = 128;
    else if (*dx <= 256)
      w = 256;
    else if (*dx <= 512)
      w = 512;
    else if (*dx <= 1024)
      w = 1024;
    else
      w = 2048;

    if (*dy <= 64)
      h = 64;
    else if (*dy <= 128)
      h = 128;
    else if (*dy <= 256)
      h = 256;
    else if (*dy <= 512)
      h = 512;
    else if (*dy <= 1024)
      h = 1024;
    else
      h = 2048;

    if (w != h) {
      if (w > h)
	h = w;
      else
	w = h;
    }

    n = malloc(w*h*4);
    if (n == NULL) {
      fprintf(stderr, "COMMON_REFORMAT_IMAGE: Out of memory error.\n");
      return FAILED;
    }

    /* clear the output buffer */
    for (i = 0; i < w*h*4; ) {
      n[i++] = editor_pen_bg_r;
      n[i++] = editor_pen_bg_g;
      n[i++] = editor_pen_bg_b;
      n[i++] = editor_pen_bg_a;
    }

    /* copy the image data */
    for (y = 0; y < h && y < *dy; y++) {
      for (x = 0; x < w && x < *dx; x++) {
	n[(y*w + x)*4 + 0] = (*d)[(y*(*dx) + x)*4 + 0];
	n[(y*w + x)*4 + 1] = (*d)[(y*(*dx) + x)*4 + 1];
	n[(y*w + x)*4 + 2] = (*d)[(y*(*dx) + x)*4 + 2];
	n[(y*w + x)*4 + 3] = (*d)[(y*(*dx) + x)*4 + 3];
      }
    }

    free(*d);
    *d = n;
    *dx = w;
    *dy = h;
  }

  return SUCCEEDED;
}


int common_get_path(char *filename, char *path) {

  int i;


  if (filename == NULL || path == NULL)
    return FAILED;

  /* strip the filename from the whole path */
  strcpy(path, filename);

  for (i = strlen(filename)-1; i >= 0; i--) {
#ifdef WIN32
    if (path[i] == '\\')
#else
    if (path[i] == '/')
#endif
      break;
  }
  path[i+1] = 0;

  return SUCCEEDED;
}


int common_scale(int i, int view_mode, int odx, int ody, unsigned char *data) {

  static unsigned char *tmp = NULL;
  static int tmp_size = 0;
  int x, y, dx, dy, a, b;
  float fr, fg, fb, fa;


  /* manage the tmp buffer */
  if (tmp_size < odx*ody*4) {
    tmp_size = odx*ody*4;
    tmp = realloc(tmp, tmp_size);
    if (tmp == NULL) {
      fprintf(stderr, "COMMON_SCALE: Out of memory error.\n");
      tmp_size = 0;
      return FAILED;
    }
  }

  if (view_mode == VIEW_MODE_RGBA) {
    for (a = 0; a < odx*ody*4; a += 4) {
      tmp[a + 0] = editor_pen_bg_r;
      tmp[a + 1] = editor_pen_bg_g;
      tmp[a + 2] = editor_pen_bg_b;
      tmp[a + 3] = editor_pen_bg_a;
    }
  }
  else if (view_mode == VIEW_MODE_RGB) {
    for (a = 0; a < odx*ody*4; a += 4) {
      tmp[a + 0] = editor_pen_bg_r;
      tmp[a + 1] = editor_pen_bg_g;
      tmp[a + 2] = editor_pen_bg_b;
      tmp[a + 3] = data[a + 3];
    }
  }
  else if (view_mode == VIEW_MODE_ALPHA) {
    for (a = 0; a < odx*ody*4; a += 4) {
      tmp[a + 0] = data[a + 0];
      tmp[a + 1] = data[a + 1];
      tmp[a + 2] = data[a + 2];
      tmp[a + 3] = editor_pen_bg_a;
    }
  }

  dx = odx / i;
  dy = ody / i;

  if (view_mode == VIEW_MODE_RGBA) {
    for (y = 0; y < dy; y++) {
      for (x = 0; x < dx; x++) {
	fr = fg = fb = fa = 0.0f;
	for (a = 0; a < i; a++) {
	  for (b = 0; b < i; b++) {
	    fr += data[(y*i+a)*odx*4 + (x*i+b)*4 + 0];
	    fg += data[(y*i+a)*odx*4 + (x*i+b)*4 + 1];
	    fb += data[(y*i+a)*odx*4 + (x*i+b)*4 + 2];
	    fa += data[(y*i+a)*odx*4 + (x*i+b)*4 + 3];
	  }
	}
	tmp[y*odx*4 + x*4 + 0] = fr/(float)(i*i);
	tmp[y*odx*4 + x*4 + 1] = fg/(float)(i*i);
	tmp[y*odx*4 + x*4 + 2] = fb/(float)(i*i);
	tmp[y*odx*4 + x*4 + 3] = fa/(float)(i*i);
      }
    }
  }
  else if (view_mode == VIEW_MODE_RGB) {
    for (y = 0; y < dy; y++) {
      for (x = 0; x < dx; x++) {
	fr = fg = fb = fa = 0.0f;
	for (a = 0; a < i; a++) {
	  for (b = 0; b < i; b++) {
	    fr += data[(y*i+a)*odx*4 + (x*i+b)*4 + 0];
	    fg += data[(y*i+a)*odx*4 + (x*i+b)*4 + 1];
	    fb += data[(y*i+a)*odx*4 + (x*i+b)*4 + 2];
	  }
	}
	tmp[y*odx*4 + x*4 + 0] = fr/(float)(i*i);
	tmp[y*odx*4 + x*4 + 1] = fg/(float)(i*i);
	tmp[y*odx*4 + x*4 + 2] = fb/(float)(i*i);
      }
    }
  }
  else if (view_mode == VIEW_MODE_ALPHA) {
    for (y = 0; y < dy; y++) {
      for (x = 0; x < dx; x++) {
	fr = fg = fb = fa = 0.0f;
	for (a = 0; a < i; a++) {
	  for (b = 0; b < i; b++) {
	    fa += data[(y*i+a)*odx*4 + (x*i+b)*4 + 3];
	  }
	}
	tmp[y*odx*4 + x*4 + 3] = fa/(float)(i*i);
      }
    }
  }

  memcpy(data, tmp, ody*odx*4);

  return SUCCEEDED;
}
