
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gtk/gtkadjustment.h>
#include <gtk/gtkwidget.h>
#include <gdk/gdkkeysyms.h>

#include "defines.h"
#include "memory.h"
#include "editor.h"
#include "common.h"
#include "palette.h"
#include "tiled.h"
#include "jpg.h"
#include "png.h"
#include "bmp.h"
#include "pcx.h"
#include "tga.h"
#include "prefs.h"
#include "string.h"


/* internal functions */
static void memory_clear(GtkWidget *widget, gpointer data);
static void memory_open(GtkWidget *widget, gpointer data);
static void memory_save(GtkWidget *widget, gpointer data);
static void memory_save_as(GtkWidget *widget, gpointer data);
static void memory_revert(GtkWidget *widget, gpointer data);
static void memory_file_save_ok(GtkWidget *widget, gpointer data);
static void memory_file_save_cancel(GtkWidget *widget, gpointer data);
static void memory_file_open_ok(GtkWidget *widget, gpointer data);
static void memory_file_open_cancel(GtkWidget *widget, gpointer data);
static void memory_size_n(GtkWidget *widget, gpointer data);
static void memory_zoom_n(GtkWidget *widget, gpointer data);
static void memory_view_n(GtkWidget *widget, gpointer data);
static void memory_scale_n(GtkWidget *widget, gpointer data);
static void memory_window_n(GtkWidget *widget, gpointer data);
static void memory_snap(GtkWidget *widget, gpointer data);
static void memory_update_mode(GtkWidget *widget, gpointer data);
static void memory_undo(GtkWidget *widget, gpointer data);
static gint memory_button_press(GtkWidget *widget, GdkEventButton *event);
static gint memory_button_release(GtkWidget *widget, GdkEventButton *event);
static gint memory_motion_notify(GtkWidget *widget, GdkEventButton *event);
static void memory_copy_to_undo(void);
static void memory_draw_rectangle_pixel(int x, int y);
static gint memory_key_press(GtkWidget *widget, GdkEventKey *event);
static gboolean memory_draw_area_expose(GtkWidget *widget, GdkEventExpose *event, gpointer user_data);
static void memory_file_read(char *n);


/* memory window data (GTK) */
GtkWidget *memory_window = NULL;
GtkWidget *memory_vbox1 = NULL;
GtkWidget *memory_menubar = NULL;
GtkWidget *memory_draw_area = NULL;
GtkWidget *memory_file_selection_save = NULL, *memory_file_selection_open = NULL;
GtkWidget *memory_scrolled_window = NULL;
GtkWidget *memory_widget_revert = NULL;

GtkAccelGroup *memory_accel_group = NULL;
GtkItemFactory *memory_item_factory = NULL;

GtkItemFactoryEntry memory_menu_items[] = {
  { "/_File",           NULL,         NULL, 0, "<Branch>" },
  { "/File/New",        NULL,         memory_clear, 0, NULL },
  { "/File/sep1",       NULL,         NULL, 0, "<Separator>" },
  { "/File/_Open",      "<control>O", memory_open, 0, NULL },
  { "/File/_Save",      "<control>S", memory_save, 0, NULL },
  { "/File/Save As",    NULL,         memory_save_as, 0, NULL },
  { "/File/Revert",     NULL,         memory_revert, 0, NULL },
  { "/File/sep2",       NULL,         NULL, 0, "<Separator>" },
  { "/File/_Quit",      "<control>Q", common_gtk_exit, 0, NULL },
  { "/_Edit",           NULL,         NULL, 0, "<Branch>" },
  { "/Edit/Undo",       "<control>Z", memory_undo, 0, NULL },
  { "/Edit/sep3",        NULL,        NULL, 0, "<Separator>" },
  { "/Edit/Scale",      NULL,         NULL, 0, "<Branch>" },
  { "/Edit/Scale/50%",  NULL,         memory_scale_n, 2, NULL },
  { "/Edit/Scale/25%",  NULL,         memory_scale_n, 4, NULL },
  { "/Edit/Scale/12.5%",NULL,         memory_scale_n, 8, NULL },
  { "/_Zoom",           NULL,         NULL, 0, "<Branch>" },
  { "/Zoom/100%",       NULL,         memory_zoom_n, 1, "<RadioItem>" },
  { "/Zoom/200%",       NULL,         memory_zoom_n, 2, "/Zoom/100%" },
  { "/Zoom/300%",       NULL,         memory_zoom_n, 3, "/Zoom/100%" },
  { "/Zoom/400%",       NULL,         memory_zoom_n, 4, "/Zoom/100%" },
  { "/_Size",           NULL,         NULL, 0, "<Branch>" },
  { "/Size/64x64",      NULL,         memory_size_n, 64, "<RadioItem>" },
  { "/Size/128x128",    NULL,         memory_size_n, 128, "/Size/64x64" },
  { "/Size/256x256",    NULL,         memory_size_n, 256, "/Size/64x64" },
  { "/Size/512x512",    NULL,         memory_size_n, 512, "/Size/64x64" },
  { "/Size/1024x1024",  NULL,         memory_size_n, 1024, "/Size/64x64" },
  { "/Size/2048x2048",  NULL,         memory_size_n, 2048, "/Size/64x64" },
  { "/_View",           NULL,         NULL, 0, "<Branch>" },
  { "/View/RGBA",       "<alt>6",     memory_view_n, VIEW_MODE_RGBA, "<RadioItem>" },
  { "/View/RGB",        "<alt>7",     memory_view_n, VIEW_MODE_RGB, "/View/RGBA" },
  { "/View/Alpha",      "<alt>8",     memory_view_n, VIEW_MODE_ALPHA, "/View/RGBA" },
  { "/View/sep4",       NULL,         NULL, 0, "<Separator>" },
  { "/View/Snap to Grid", NULL,       memory_snap, 0, "<ToggleItem>" },
  { "/View/sep5",       NULL,         NULL, 0, "<Separator>" },
  { "/View/Updated in Real-Time",     NULL, memory_update_mode, 0, "<ToggleItem>" },
  { "/_Windows",        NULL,         NULL, 0, "<Branch>" },
  { "/Windows/Tiled View", NULL,      memory_window_n, 0, "<ToggleItem>" },
  { "/Windows/Palette", NULL,         memory_window_n, 1, "<ToggleItem>" },
};

/* memory data */
int memory_zoom = 2;
int memory_dx = 128, memory_dy = 128;
unsigned char *memory_data = NULL, *memory_view = NULL;
int memory_view_mode = VIEW_MODE_RGBA;

/* file data */
char memory_file_name[256], memory_file_path[256];
int memory_file_status = NO, memory_file_name_status = NO;

/* mouse */
int memory_mouse_button = OFF;
int memory_x = 0, memory_y = 0;
int memory_snap_to_grid = NO;

/* real time updates */
int memory_real_time = NO;

/* undo */
unsigned char *memory_undo_data = NULL, *memory_undo_tmp = NULL;
int memory_undo_dx = 0, memory_undo_dy = 0;



int memory_window_init(void) {

  /* window */
  memory_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  memory_window_set_title();

  /* menu */
  memory_accel_group = gtk_accel_group_new();
  memory_item_factory = gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<main>", memory_accel_group);
  gtk_item_factory_create_items(memory_item_factory, sizeof(memory_menu_items)/sizeof(memory_menu_items[0]), memory_menu_items, NULL);
  gtk_window_add_accel_group(GTK_WINDOW(memory_window), memory_accel_group);
  memory_menubar = gtk_item_factory_get_widget(memory_item_factory, "<main>");

  /* drawing area */
  memory_draw_area = gtk_drawing_area_new();
  gtk_drawing_area_size(GTK_DRAWING_AREA(memory_draw_area), memory_dx*memory_zoom, memory_dy*memory_zoom);
  gtk_widget_add_events(memory_draw_area,
			GDK_KEY_PRESS_MASK | GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK | GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK | GDK_POINTER_MOTION_MASK);

  /* scrolled window */
  memory_scrolled_window = gtk_scrolled_window_new(NULL, NULL);
  gtk_container_set_border_width(GTK_CONTAINER(memory_scrolled_window), 4);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(memory_scrolled_window),
				 GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

  /* boxes */
  memory_vbox1 = gtk_vbox_new(FALSE, 0);
  gtk_container_border_width(GTK_CONTAINER(memory_vbox1), 1);

  /* place boxes and widgets */
  gtk_container_add(GTK_CONTAINER(memory_window), memory_vbox1);
  gtk_box_pack_start(GTK_BOX(memory_vbox1), memory_menubar, FALSE, FALSE, 0);
  gtk_container_add(GTK_CONTAINER(memory_vbox1), memory_scrolled_window);
  gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(memory_scrolled_window), memory_draw_area);

  /* file selections */
  memory_file_selection_save = gtk_file_selection_new("Save PNG Image");
  memory_file_selection_open = gtk_file_selection_new("Open BMP/JPG/PCX/PNG/TGA Image");

  /* signals */
  gtk_signal_connect(GTK_OBJECT(memory_draw_area), "expose_event", GTK_SIGNAL_FUNC(memory_draw_area_expose), NULL);
  gtk_signal_connect(GTK_OBJECT(memory_draw_area), "button_press_event", GTK_SIGNAL_FUNC(memory_button_press), NULL);
  gtk_signal_connect(GTK_OBJECT(memory_draw_area), "button_release_event", GTK_SIGNAL_FUNC(memory_button_release), NULL);
  gtk_signal_connect(GTK_OBJECT(memory_draw_area), "motion_notify_event", GTK_SIGNAL_FUNC(memory_motion_notify), NULL);
  gtk_signal_connect(GTK_OBJECT(memory_window), "key_press_event", GTK_SIGNAL_FUNC(memory_key_press), NULL);
  gtk_signal_connect(GTK_OBJECT(memory_window), "delete_event", GTK_SIGNAL_FUNC(common_gtk_exit), NULL);
  gtk_signal_connect(GTK_OBJECT(memory_window), "destroy", GTK_SIGNAL_FUNC(common_gtk_exit), NULL);

  gtk_signal_connect(GTK_OBJECT(memory_file_selection_save), "delete_event", GTK_SIGNAL_FUNC(common_gtk_hide_widget), NULL);
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(memory_file_selection_save)->ok_button), "clicked", GTK_SIGNAL_FUNC(memory_file_save_ok), NULL);
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(memory_file_selection_save)->cancel_button), "clicked", GTK_SIGNAL_FUNC(memory_file_save_cancel), NULL);

  gtk_signal_connect(GTK_OBJECT(memory_file_selection_open), "delete_event", GTK_SIGNAL_FUNC(common_gtk_hide_widget), NULL);
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(memory_file_selection_open)->ok_button), "clicked", GTK_SIGNAL_FUNC(memory_file_open_ok), NULL);
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(memory_file_selection_open)->cancel_button), "clicked", GTK_SIGNAL_FUNC(memory_file_open_cancel), NULL);

  gtk_window_set_resizable(GTK_WINDOW(memory_window), TRUE);
  gtk_widget_set_size_request(memory_window, 256, 256);

  memory_file_name[0] = 0;

  /* get some widgets */
  memory_widget_revert = gtk_item_factory_get_widget(memory_item_factory, "/File/Revert");

  return SUCCEEDED;
}


static gint memory_key_press(GtkWidget *widget, GdkEventKey *event) {

  if (event->keyval == GDK_Right) {
    /* move one tile right */
    if (memory_x < memory_dx - editor_dx) {
      memory_x += editor_dx;
      if (memory_compute_tile_position(memory_x*memory_zoom, memory_y*memory_zoom) == SUCCEEDED) {
	editor_window_set_title();
	memory_copy_data_out(editor_data, memory_x, memory_y, editor_dx, editor_dy);
	editor_window_refresh();
	memory_window_refresh();
	tiled_window_refresh();
      }
    }
  }
  else if (event->keyval == GDK_Left) {
    /* move one tile left */
    if (memory_x >= editor_dx) {
      memory_x -= editor_dx;
      if (memory_compute_tile_position(memory_x*memory_zoom, memory_y*memory_zoom) == SUCCEEDED) {
	editor_window_set_title();
	memory_copy_data_out(editor_data, memory_x, memory_y, editor_dx, editor_dy);
	editor_window_refresh();
	memory_window_refresh();
	tiled_window_refresh();
      }
    }
  }
  else if (event->keyval == GDK_Up) {
    /* move one tile up */
    if (memory_y >= editor_dy) {
      memory_y -= editor_dy;
      if (memory_compute_tile_position(memory_x*memory_zoom, memory_y*memory_zoom) == SUCCEEDED) {
	editor_window_set_title();
	memory_copy_data_out(editor_data, memory_x, memory_y, editor_dx, editor_dy);
	editor_window_refresh();
	memory_window_refresh();
	tiled_window_refresh();
      }
    }
  }
  else if (event->keyval == GDK_Down) {
    /* move one tile down */
    if (memory_y < memory_dy - editor_dy) {
      memory_y += editor_dy;
      if (memory_compute_tile_position(memory_x*memory_zoom, memory_y*memory_zoom) == SUCCEEDED) {
	editor_window_set_title();
	memory_copy_data_out(editor_data, memory_x, memory_y, editor_dx, editor_dy);
	editor_window_refresh();
	memory_window_refresh();
	tiled_window_refresh();
      }
    }
  }

  return FALSE;
}


int memory_copy_data_out(unsigned char *out, int x, int y, int dx, int dy) {

  int a, b, i, n, p;


  if (out == NULL || dx <= 0 || dy <= 0)
    return FAILED;

  for (i = 0, a = 0; a < dy; a++) {
    p = (y+a)*memory_dx + x;
    for (b = 0; b < dx; b++) {
      n = (p + b)<<2;
      out[i++] = memory_data[n + 0];
      out[i++] = memory_data[n + 1];
      out[i++] = memory_data[n + 2];
      out[i++] = memory_data[n + 3];
    }
  }

  return SUCCEEDED;
}


int memory_copy_data_in(unsigned char *in, int x, int y, int dx, int dy) {

  int a, b, i, n, p;


  if (in == NULL || dx <= 0 || dy <= 0)
    return FAILED;

  for (i = 0, a = 0; a < dy; a++) {
    p = (y+a)*memory_dx + x;
    for (b = 0; b < dx; b++) {
      n = (p + b)<<2;
      memory_data[n + 0] = in[i++];
      memory_data[n + 1] = in[i++];
      memory_data[n + 2] = in[i++];
      memory_data[n + 3] = in[i++];
    }
  }

  return SUCCEEDED;
}


static void memory_draw_rectangle_pixel(int x, int y) {

  unsigned char *d;
  int z, a, b, dx, dy, q;


  d = memory_view;
  z = memory_zoom;
  dx = memory_dx;
  dy = memory_dy;

  for (a = 0; a < z; a++) {
    for (b = 0; b < z; b++) {
      q = 255 - d[(y*z*dx*z + x*z + a*dx*z + b)*3 + 0];
      d[(y*z*dx*z + x*z + a*dx*z + b)*3 + 0] = q;
      q = 255 - d[(y*z*dx*z + x*z + a*dx*z + b)*3 + 1];
      d[(y*z*dx*z + x*z + a*dx*z + b)*3 + 1] = q;
      q = 255 - d[(y*z*dx*z + x*z + a*dx*z + b)*3 + 2];
      d[(y*z*dx*z + x*z + a*dx*z + b)*3 + 2] = q;
    }
  }
}


int memory_draw_editor_rectangle(void) {

  int x, y;


  /* top */
  for (x = 0; x < editor_dx; x++)
    memory_draw_rectangle_pixel(memory_x + x, memory_y);

  /* bottom */
  for (x = 0; x < editor_dx; x++)
    memory_draw_rectangle_pixel(memory_x + x, memory_y + editor_dy-1);

  /* left */
  for (y = 1; y < editor_dy-1; y++)
    memory_draw_rectangle_pixel(memory_x, memory_y + y);

  /* right */
  for (y = 1; y < editor_dy-1; y++)
    memory_draw_rectangle_pixel(memory_x + editor_dx-1, memory_y + y);

  return SUCCEEDED;
}


void memory_title_reset(int ents) {

  if (editor_need_to_save == ents)
    return;

  editor_need_to_save = ents;
  memory_window_set_title();
}


void memory_window_set_title(void) {

  char tmp[512];
  int i;


  if (memory_window == NULL)
    return;

  if (memory_file_status == YES) {
    for (i = strlen(memory_file_name)-1; i >= 0; i--) {
#ifdef WIN32
      if (memory_file_name[i] == '\\')
#else
      if (memory_file_name[i] == '/')
#endif
	break;
    }

    if (editor_need_to_save == NO) {
      sprintf(tmp, "%s | %s | %dx%d | ", version_string, &memory_file_name[i+1], memory_dx, memory_dy);
      if (memory_widget_revert != NULL)
	gtk_widget_set_sensitive(memory_widget_revert, FALSE);
    }
    else {
      sprintf(tmp, "%s | *%s | %dx%d | ", version_string, &memory_file_name[i+1], memory_dx, memory_dy);
      if (memory_widget_revert != NULL)
	gtk_widget_set_sensitive(memory_widget_revert, TRUE);
    }
  }
  else {
    sprintf(tmp, "%s | ? | %dx%d | ", version_string, memory_dx, memory_dy);
    if (memory_widget_revert != NULL)
      gtk_widget_set_sensitive(memory_widget_revert, FALSE);
  }

  if (memory_view_mode == VIEW_MODE_RGBA)
    strcat(tmp, "RGBA");
  else if (memory_view_mode == VIEW_MODE_RGB)
    strcat(tmp, "RGB");
  else if (memory_view_mode == VIEW_MODE_ALPHA)
    strcat(tmp, "Alpha");

  gtk_window_set_title(GTK_WINDOW(memory_window), tmp);
}


void memory_resize_view(int zoom) {

  memory_view = realloc(memory_view, memory_dx*zoom*memory_dy*zoom*3);
  if (memory_view == NULL) {
    fprintf(stderr, "MEMORY_RESIZE_VIEW: Out of memory error.\n");
    return;
  }

  memory_zoom = zoom;
}


void memory_resize_data(int dx, int dy) {

  unsigned char *m;
  int i, ox, oy, x, y;


  /* save the old */
  m = memory_data;
  ox = memory_dx;
  oy = memory_dy;

  memory_data = malloc(dx*dy*4);
  if (memory_data == NULL) {
    fprintf(stderr, "MEMORY_RESIZE_DATA: Out of memory error.\n");
    free(m);
    return;
  }

  memory_dx = dx;
  memory_dy = dy;

  /* clear the new image */
  for (i = 0; i < memory_dx*memory_dy*4; i += 4) {
    memory_data[i + 0] = editor_pen_bg_r;
    memory_data[i + 1] = editor_pen_bg_g;
    memory_data[i + 2] = editor_pen_bg_b;
    memory_data[i + 3] = editor_pen_bg_a;
  }

  /* copy as much from the old image as possible */
  if (m != NULL) {
    for (y = 0; y < oy && y < dy; y++) {
      for (x = 0; x < ox && x < dx; x++) {
	memory_data[(y*dx + x)*4 + 0] = m[(y*ox + x)*4 + 0];
	memory_data[(y*dx + x)*4 + 1] = m[(y*ox + x)*4 + 1];
	memory_data[(y*dx + x)*4 + 2] = m[(y*ox + x)*4 + 2];
	memory_data[(y*dx + x)*4 + 3] = m[(y*ox + x)*4 + 3];
      }
    }

    free(m);
  }
}


static void memory_clear(GtkWidget *widget, gpointer data) {

  int i;


  /* backup the old data */
  memory_copy_to_undo();

  memory_title_reset(YES);

  /* clear the data */
  for (i = 0; i < memory_dx*memory_dy*4; i += 4) {
    memory_data[i + 0] = editor_pen_bg_r;
    memory_data[i + 1] = editor_pen_bg_g;
    memory_data[i + 2] = editor_pen_bg_b;
    memory_data[i + 3] = editor_pen_bg_a;
  }

  common_copy_data_to_view(memory_data, memory_view, memory_dx, memory_dy, memory_zoom, memory_view_mode);
  memory_draw_editor_rectangle();

  /* update the window title */
  memory_file_status = NO;
  memory_file_name_status = NO;
  memory_window_set_title();

  /* refresh the memory draw area */
  gtk_widget_queue_draw(memory_draw_area);

  /* refresh the windows */
  memory_copy_data_out(editor_data, memory_x, memory_y, editor_dx, editor_dy);
  editor_window_refresh();
  tiled_window_refresh();
}


static void memory_scale_n(GtkWidget *widget, gpointer data) {

  /* backup the old data */
  memory_copy_to_undo();

  memory_title_reset(YES);

  /* scale! */
  common_scale((int)data, memory_view_mode, memory_dx, memory_dy, memory_data);

  /* refresh the memory draw area */
  common_copy_data_to_view(memory_data, memory_view, memory_dx, memory_dy, memory_zoom, memory_view_mode);
  memory_draw_editor_rectangle();
  gtk_widget_queue_draw(memory_draw_area);

  /* refresh the windows */
  memory_copy_data_out(editor_data, memory_x, memory_y, editor_dx, editor_dy);
  editor_window_refresh();
  tiled_window_refresh();
}


static void memory_copy_to_undo(void) {

  if (memory_dx != memory_undo_dx || memory_dy != memory_undo_dy) {
    memory_undo_data = realloc(memory_undo_data, memory_dx*memory_dy*4);
    memory_undo_tmp = realloc(memory_undo_tmp, memory_dx*memory_dy*4);
    if (memory_undo_data == NULL || memory_undo_tmp == NULL) {
      fprintf(stderr, "MEMORY_COPY_TO_UNDO: Out of memory error.\n");
      return;
    }

    memory_undo_dx = memory_dx;
    memory_undo_dy = memory_dy;
  }

  /* backup the data */
  memcpy(memory_undo_data, memory_data, memory_dx*memory_dy*4);
}


static void memory_undo(GtkWidget *widget, gpointer data) {

  if (memory_undo_dx != memory_dx || memory_undo_dy != memory_dy || memory_undo_data == NULL || memory_undo_tmp == NULL)
    return;

  /* backup the current data */
  memcpy(memory_undo_tmp, memory_data, memory_dx*memory_dy*4);

  /* return the old data */
  memcpy(memory_data, memory_undo_data, memory_dx*memory_dy*4);

  /* store the previous data */
  memcpy(memory_undo_data, memory_undo_tmp, memory_dx*memory_dy*4);

  /* refresh the memory draw area */
  common_copy_data_to_view(memory_data, memory_view, memory_dx, memory_dy, memory_zoom, memory_view_mode);
  memory_draw_editor_rectangle();
  gtk_widget_queue_draw(memory_draw_area);

  /* refresh the windows */
  memory_copy_data_out(editor_data, memory_x, memory_y, editor_dx, editor_dy);
  editor_window_refresh();
  tiled_window_refresh();
}


int memory_window_refresh(void) {

  if (memory_data == NULL || memory_view == NULL)
    return FAILED;

  common_copy_data_to_view(memory_data, memory_view, memory_dx, memory_dy, memory_zoom, memory_view_mode);
  memory_draw_editor_rectangle();

  /* refresh the memory draw area */
  gtk_widget_queue_draw(memory_draw_area);

  return SUCCEEDED;
}


int memory_window_block_refresh(void) {

  if (memory_data == NULL || memory_view == NULL)
    return FAILED;

  common_copy_block_to_view(memory_data, memory_view, memory_dx, memory_dy, memory_zoom, memory_view_mode, memory_x, memory_y, editor_dx, editor_dy);
  memory_draw_editor_rectangle();

  /* refresh the memory draw area */
  gdk_draw_rgb_image(memory_draw_area->window, memory_draw_area->style->fg_gc[GTK_STATE_NORMAL], memory_x*memory_zoom, memory_y*memory_zoom,
		     editor_dx*memory_zoom, editor_dy*memory_zoom, GDK_RGB_DITHER_MAX, memory_view + memory_y*memory_zoom*memory_dx*3*memory_zoom + memory_x*3*memory_zoom, memory_dx*memory_zoom*3);

  return SUCCEEDED;
}


int memory_window_plain_block_refresh(void) {

  if (memory_data == NULL || memory_view == NULL)
    return FAILED;

  common_copy_block_to_view(memory_data, memory_view, memory_dx, memory_dy, memory_zoom, memory_view_mode, memory_x, memory_y, editor_dx, editor_dy);

  /* refresh the memory draw area */
  gdk_draw_rgb_image(memory_draw_area->window, memory_draw_area->style->fg_gc[GTK_STATE_NORMAL], memory_x*memory_zoom, memory_y*memory_zoom,
		     editor_dx*memory_zoom, editor_dy*memory_zoom, GDK_RGB_DITHER_MAX, memory_view + memory_y*memory_zoom*memory_dx*3*memory_zoom + memory_x*3*memory_zoom, memory_dx*memory_zoom*3);

  return SUCCEEDED;
}


static void memory_save(GtkWidget *widget, gpointer data) {

  if (memory_file_name_status == NO) {
    memory_save_as(widget, data);
    return;
  }

  png_save(memory_file_name, memory_dx, memory_dy, memory_data);

  editor_need_to_save = NO;
  memory_window_set_title();
}


static void memory_save_as(GtkWidget *widget, gpointer data) {

  gtk_widget_show(memory_file_selection_save);
}


static void memory_open(GtkWidget *widget, gpointer data) {

  gtk_widget_show(memory_file_selection_open);
}


static void memory_file_save_ok(GtkWidget *widget, gpointer data) {

  char tmp[512];
  gchar *n;


  editor_need_to_save = NO;
  memory_window_set_title();

  n = (gchar *)gtk_file_selection_get_filename(GTK_FILE_SELECTION(memory_file_selection_save));
  gtk_file_selection_set_filename(GTK_FILE_SELECTION(memory_file_selection_open), n);

  /* hide the file selection window */
  gtk_widget_hide(memory_file_selection_save);

  png_save(n, memory_dx, memory_dy, memory_data);

  strcpy(memory_file_name, n);
  memory_file_status = YES;
  memory_file_name_status = YES;

  /* strip the filename from the whole path */
  common_get_path(n, tmp);
  strcpy(memory_file_path, tmp);

  memory_window_set_title();
}


static void memory_revert(GtkWidget *widget, gpointer data) {

  if (memory_file_status == NO)
    return;

  /* backup the old data */
  memory_copy_to_undo();

  memory_file_read(memory_file_name);
}


static void memory_file_open_ok(GtkWidget *widget, gpointer data) {

  gchar *n;


  /* backup the old data */
  memory_copy_to_undo();

  n = (gchar *)gtk_file_selection_get_filename(GTK_FILE_SELECTION(memory_file_selection_open));

  /* hide the file selection window */
  gtk_widget_hide(memory_file_selection_open);

  memory_file_read((char *)n);
}


static void memory_file_read(char *n) {

  unsigned char *d;
  int dx, dy, bpp, i, l;
  char tmp[512];


  /* find the file type */
  for (i = strlen(n)-1; i >= 0; i--) {
    if (n[i] == '.')
      break;
  }

  l = 0;
  if (string_compare_nocase(&n[i+1], "png") == 0)
    l |= 1;
  else if (string_compare_nocase(&n[i+1], "jpg") == 0 || string_compare_nocase(&n[i+1], "jpeg") == 0)
    l |= 2;
  else if (string_compare_nocase(&n[i+1], "bmp") == 0)
    l |= 4;
  else if (string_compare_nocase(&n[i+1], "pcx") == 0)
    l |= 8;
  else if (string_compare_nocase(&n[i+1], "tga") == 0)
    l |= 16;
  else
    l |= 31;

  i = FAILED;
  if (l & 1)
    i = png_load(n, &dx, &dy, &bpp, &d);
  if (l & 2 && i == FAILED)
    i = jpg_load(n, &dx, &dy, &bpp, &d);
  if (l & 4 && i == FAILED)
    i = bmp_load(n, &dx, &dy, &bpp, &d);
  if (l & 8 && i == FAILED)
    i = pcx_load(n, &dx, &dy, &bpp, &d);
  if (l & 16 && i == FAILED)
    i = tga_load(n, &dx, &dy, &bpp, &d);

  if (i == FAILED)
    return;

  /* reformat the image so it is of suitable format */
  common_reformat_image(&dx, &dy, &bpp, &d);

  strcpy(memory_file_name, n);
  memory_file_status = YES;

  /* strip the filename from the whole path */
  common_get_path(n, tmp);
  strcpy(memory_file_path, tmp);

  /* the filename can be used later only if the file was a png file */
  if (l == 1) {
    memory_file_name_status = YES;

    /* set the whole filename with path to the "save as" file selector */
    gtk_file_selection_set_filename(GTK_FILE_SELECTION(memory_file_selection_save), n);
  }
  else {
    memory_file_name_status = NO;

    /* set the path to the "save as" file selector */
    gtk_file_selection_set_filename(GTK_FILE_SELECTION(memory_file_selection_save), tmp);
  }

  if (dx != memory_dx) {
    if (dx == 64)
      gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(memory_item_factory, "/Size/64x64"), TRUE);
    else if (dx == 128)
      gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(memory_item_factory, "/Size/128x128"), TRUE);
    else if (dx == 256)
      gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(memory_item_factory, "/Size/256x256"), TRUE);
    else if (dx == 512)
      gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(memory_item_factory, "/Size/512x512"), TRUE);
    else if (dx == 1024)
      gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(memory_item_factory, "/Size/1024x1024"), TRUE);
    else
      gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(memory_item_factory, "/Size/2048x2048"), TRUE);

    memory_dx = dx;
    memory_dy = dy;
  }

  if (memory_data != NULL)
    free(memory_data);
  memory_data = d;

  memory_resize_view(memory_zoom);
  common_copy_data_to_view(memory_data, memory_view, memory_dx, memory_dy, memory_zoom, memory_view_mode);
  memory_draw_editor_rectangle();

  gtk_drawing_area_size(GTK_DRAWING_AREA(memory_draw_area), memory_dx*memory_zoom, memory_dy*memory_zoom);

  /* refresh the windows */
  memory_copy_data_out(editor_data, memory_x, memory_y, editor_dx, editor_dy);
  editor_window_refresh();
  tiled_window_refresh();

  editor_need_to_save = NO;
  memory_window_set_title();
}


static void memory_file_save_cancel(GtkWidget *widget, gpointer data) {

  gtk_widget_hide(memory_file_selection_save);
}


static void memory_file_open_cancel(GtkWidget *widget, gpointer data) {

  gtk_widget_hide(memory_file_selection_open);
}


static void memory_zoom_n(GtkWidget *widget, gpointer data) {

  int zoom;


  zoom = (int)data;
  if (memory_zoom == zoom)
    return;

  memory_resize_view(zoom);
  common_copy_data_to_view(memory_data, memory_view, memory_dx, memory_dy, memory_zoom, memory_view_mode);
  memory_draw_editor_rectangle();

  gtk_drawing_area_size(GTK_DRAWING_AREA(memory_draw_area), memory_dx*memory_zoom, memory_dy*memory_zoom);
}


static void memory_size_n(GtkWidget *widget, gpointer data) {

  int i;


  i = (int)data;
  if (memory_dx == i && memory_dy == i)
    return;

  /* memory window cannot be smaller than the tile size */
  if (i < editor_dx || i < editor_dy) {
    if (memory_dx == 64)
      gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(memory_item_factory, "/Size/64x64"), TRUE);
    else if (memory_dx == 128)
      gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(memory_item_factory, "/Size/128x128"), TRUE);
    else if (memory_dx == 256)
      gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(memory_item_factory, "/Size/256x256"), TRUE);
    else if (memory_dx == 512)
      gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(memory_item_factory, "/Size/512x512"), TRUE);
    else if (memory_dx == 1024)
      gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(memory_item_factory, "/Size/1024x1024"), TRUE);
    else if (memory_dx == 2048)
      gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(memory_item_factory, "/Size/2048x2048"), TRUE);
    return;
  }

  /* move the tile pointer to (0, 0) */
  memory_compute_tile_position(0, 0);
  editor_window_set_title();

  memory_resize_data(i, i);
  memory_resize_view(memory_zoom);
  common_copy_data_to_view(memory_data, memory_view, memory_dx, memory_dy, memory_zoom, memory_view_mode);
  memory_draw_editor_rectangle();

  gtk_drawing_area_size(GTK_DRAWING_AREA(memory_draw_area), memory_dx*memory_zoom, memory_dy*memory_zoom);

  /* refresh the windows */
  memory_copy_data_out(editor_data, memory_x, memory_y, editor_dx, editor_dy);
  editor_window_refresh();
  tiled_window_refresh();

  editor_need_to_save = YES;
  memory_window_set_title();
}


static void memory_view_n(GtkWidget *widget, gpointer data) {

  int i;


  i = (int)data;
  if (memory_view_mode == i)
    return;

  memory_view_mode = i;
  memory_window_refresh();

  memory_window_set_title();
}


int memory_compute_tile_position(int mx, int my) {

  int x, y, i;


  /* get the position on the 1:1 image */
  x = mx/memory_zoom;
  y = my/memory_zoom;

  if (x < 0)
    x = 0;
  if (y < 0)
    y = 0;
  if (x > memory_dx - 1)
    x = memory_dx - 1;
  if (y > memory_dy - 1)
    y = memory_dy - 1;

  /* compute the tile */
  x /= editor_dx;
  y /= editor_dy;
  i = y*(memory_dx/editor_dx) + x;

  if (editor_tile == i && memory_snap_to_grid == YES)
    return FAILED;

  /* we moved to a new tile */
  editor_tile = i;
  memory_x = x*editor_dx;
  memory_y = y*editor_dy;

  return SUCCEEDED;
}


static gint memory_button_press(GtkWidget *widget, GdkEventButton *event) {

  int x, y, a, b;


  memory_mouse_button = ON;

  if (memory_snap_to_grid == YES) {
    /* snap to grid */
    x = memory_x;
    y = memory_y;
    memory_compute_tile_position(event->x, event->y);

    if (memory_x != x || memory_y != y) {
      a = memory_x;
      b = memory_y;
      memory_x = x;
      memory_y = y;
      memory_window_plain_block_refresh();
      memory_x = a;
      memory_y = b;
      memory_window_block_refresh();
    }
  }
  else {
    /* free movement of the editor rectangle */
    x = (int)event->x/memory_zoom;
    y = (int)event->y/memory_zoom;

    if (x < 0)
      x = 0;
    if (y < 0)
      y = 0;
    if (x >= memory_dx - editor_dx)
      x = memory_dx - editor_dx;
    if (y >= memory_dy - editor_dy)
      y = memory_dy - editor_dy;

    if (x != memory_x || y != memory_y) {
      memory_window_plain_block_refresh();
      memory_x = x;
      memory_y = y;
      memory_window_block_refresh();
    }
  }

  return FALSE;
}


static gint memory_button_release(GtkWidget *widget, GdkEventButton *event) {

  memory_mouse_button = OFF;

  editor_window_set_title();
  memory_copy_data_out(editor_data, memory_x, memory_y, editor_dx, editor_dy);
  editor_window_refresh();
  tiled_window_refresh();

  return FALSE;
}


static gint memory_motion_notify(GtkWidget *widget, GdkEventButton *event) {

  int x, y, a, b;


  if (memory_mouse_button == OFF)
    return FALSE;

  if (memory_snap_to_grid == YES) {
    /* snap to grid */
    x = memory_x;
    y = memory_y;
    memory_compute_tile_position(event->x, event->y);

    if (memory_x != x || memory_y != y) {
      a = memory_x;
      b = memory_y;
      memory_x = x;
      memory_y = y;
      memory_window_plain_block_refresh();
      memory_x = a;
      memory_y = b;
      memory_window_block_refresh();
    }
  }
  else {
    /* free movement of the editor rectangle */
    x = (int)event->x/memory_zoom;
    y = (int)event->y/memory_zoom;

    if (x < 0)
      x = 0;
    if (y < 0)
      y = 0;
    if (x >= memory_dx - editor_dx)
      x = memory_dx - editor_dx;
    if (y >= memory_dy - editor_dy)
      y = memory_dy - editor_dy;

    if (x != memory_x || y != memory_y) {
      memory_window_plain_block_refresh();
      memory_x = x;
      memory_y = y;
      memory_window_block_refresh();
    }
  }

  return FALSE;
}


static void memory_snap(GtkWidget *widget, gpointer data) {

  if (memory_snap_to_grid == YES)
    memory_snap_to_grid = NO;
  else
    memory_snap_to_grid = YES;

  editor_tile = -1;

  if (memory_snap_to_grid == YES) {
    /* adjust the coordinates */
    memory_compute_tile_position(memory_x*memory_zoom, memory_y*memory_zoom);

    memory_window_refresh();
    memory_copy_data_out(editor_data, memory_x, memory_y, editor_dx, editor_dy);
    editor_window_refresh();
    tiled_window_refresh();
  }

  editor_window_set_title();
}


static void memory_window_n(GtkWidget *widget, gpointer data) {

  int i;


  i = (int)data;
  if (i == 0) {
    /* tiled view */
    if (tiled_window == NULL)
      return;

    if (tiled_view_status == ON) {
      /* before hiding the window, get its origin */
      gtk_window_get_position(GTK_WINDOW(tiled_window), &tiled_window_x, &tiled_window_y);
      tiled_view_status = OFF;
      gtk_widget_hide(tiled_window);
    }
    else {
      tiled_view_status = ON;
      gtk_widget_show_all(tiled_window);
      tiled_window_refresh();
    }
  }
  else if (i == 1) {
    /* palette */
    if (palette_window == NULL)
      return;

    if (palette_status == ON) {
      /* before hiding the window, get its origin */
      gtk_window_get_position(GTK_WINDOW(palette_window), &palette_window_x, &palette_window_y);
      palette_status = OFF;
      gtk_widget_hide(palette_window);
    }
    else {
      palette_status = ON;
      gtk_widget_show_all(palette_window);
    }
  }
}


static void memory_update_mode(GtkWidget *widget, gpointer data) {

  if (memory_real_time == NO)
    memory_real_time = YES;
  else
    memory_real_time = NO;
}


static gboolean memory_draw_area_expose(GtkWidget *widget, GdkEventExpose *event, gpointer user_data) {

  if (memory_view != NULL) {
    gdk_draw_rgb_image(widget->window, widget->style->fg_gc[GTK_STATE_NORMAL], 0, 0,
		       memory_dx*memory_zoom, memory_dy*memory_zoom, GDK_RGB_DITHER_MAX, memory_view, memory_dx*memory_zoom*3);
  }

  return FALSE;
}
