
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gtk/gtkadjustment.h>
#include <gtk/gtkwidget.h>
#include <gdk/gdkkeysyms.h>

#include "defines.h"
#include "memory.h"
#include "editor.h"
#include "common.h"
#include "palette.h"
#include "png.h"


/* internal functions */
static void palette_fg_button_clicked(void);
static void palette_bg_button_clicked(void);
static void palette_selection_change(GtkWidget *widget, gpointer data);
static gint palette_hide_window(GtkWidget *widget, gpointer data);


/* palette window data (GTK) */
GtkWidget *palette_window = NULL;
GtkWidget *palette_vbox1 = NULL;
GtkWidget *palette_mode_hbox1 = NULL;
GtkWidget *palette_selection = NULL;
GtkWidget *palette_fg_button = NULL;
GtkWidget *palette_bg_button = NULL;
GSList *palette_rgroup = NULL;

int palette_status = ON;

static int palette_select_mode = FG;


int palette_window_init(void) {

  char tmp[512];


  /* window */
  palette_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  sprintf(tmp, "Color Selection");
  gtk_window_set_title(GTK_WINDOW(palette_window), tmp);

  /* boxes */
  palette_vbox1 = gtk_vbox_new(FALSE, 0);
  gtk_container_border_width(GTK_CONTAINER(palette_vbox1), 4);
  
  palette_mode_hbox1 = gtk_hbox_new(FALSE, 0);
  gtk_container_border_width(GTK_CONTAINER(palette_mode_hbox1), 4);

  palette_fg_button = gtk_radio_button_new_with_label(palette_rgroup, "Foreground");
  palette_rgroup = gtk_radio_button_group (GTK_RADIO_BUTTON(palette_fg_button));
  palette_bg_button = gtk_radio_button_new_with_label(palette_rgroup, "Background");

  /* color selection widget */
  palette_selection = gtk_color_selection_new();
  gtk_color_selection_set_has_opacity_control(GTK_COLOR_SELECTION(palette_selection), TRUE);
  gtk_color_selection_set_update_policy(GTK_COLOR_SELECTION(palette_selection), GTK_UPDATE_DISCONTINUOUS);

  /* place boxes and widgets */
  gtk_container_add(GTK_CONTAINER(palette_window), palette_vbox1);
  gtk_box_pack_start(GTK_BOX(palette_vbox1), palette_selection, FALSE, FALSE, 0);
  gtk_container_add(GTK_CONTAINER(palette_vbox1), palette_mode_hbox1);
  
  gtk_box_pack_start(GTK_BOX(palette_mode_hbox1), palette_fg_button, TRUE, TRUE, 0);
  gtk_container_add(GTK_CONTAINER(palette_mode_hbox1), palette_bg_button);

  /* signals */
  gtk_signal_connect(GTK_OBJECT(palette_window), "delete_event", GTK_SIGNAL_FUNC(palette_hide_window), NULL);
  gtk_signal_connect(GTK_OBJECT(palette_window), "destroy", GTK_SIGNAL_FUNC(palette_hide_window), NULL);
  gtk_signal_connect(GTK_OBJECT(palette_selection), "color_changed", GTK_SIGNAL_FUNC(palette_selection_change), NULL);
  
  gtk_signal_connect(GTK_OBJECT(palette_fg_button), "clicked", GTK_SIGNAL_FUNC(palette_fg_button_clicked), NULL);
  gtk_signal_connect(GTK_OBJECT(palette_bg_button), "clicked", GTK_SIGNAL_FUNC(palette_bg_button_clicked), NULL);

  gtk_window_set_resizable(GTK_WINDOW(palette_window), FALSE);

  /* use the default color */
  palette_window_refresh();

  return SUCCEEDED;
}


static void palette_fg_button_clicked(void) {

  palette_select_mode = FG;
  palette_window_refresh();
}


static void palette_bg_button_clicked(void) {

  palette_select_mode = BG;
  palette_window_refresh();
}



static void palette_selection_change(GtkWidget *widget, gpointer data) {

  gdouble c[4];


  gtk_color_selection_get_color(GTK_COLOR_SELECTION(palette_selection), c);
  
  if (palette_select_mode == FG) {
    editor_pen_r = (unsigned char)(c[0]*255+0.00001);
    editor_pen_g = (unsigned char)(c[1]*255+0.00001);
    editor_pen_b = (unsigned char)(c[2]*255+0.00001);
    editor_pen_a = (unsigned char)(c[3]*255+0.00001);
  }
  else {
    editor_pen_bg_r = (unsigned char)(c[0]*255+0.00001);
    editor_pen_bg_g = (unsigned char)(c[1]*255+0.00001);
    editor_pen_bg_b = (unsigned char)(c[2]*255+0.00001);
    editor_pen_bg_a = (unsigned char)(c[3]*255+0.00001);
  }
}


int palette_window_refresh(void) {

  gdouble c[4];

  if (palette_select_mode == FG) {
    c[0] = ((gdouble)editor_pen_r)/255.0;
    c[1] = ((gdouble)editor_pen_g)/255.0;
    c[2] = ((gdouble)editor_pen_b)/255.0;
    c[3] = ((gdouble)editor_pen_a)/255.0;
  }
  else {
    c[0] = ((gdouble)editor_pen_bg_r)/255.0;
    c[1] = ((gdouble)editor_pen_bg_g)/255.0;
    c[2] = ((gdouble)editor_pen_bg_b)/255.0;
    c[3] = ((gdouble)editor_pen_bg_a)/255.0;	  
  }
  
  gtk_color_selection_set_color(GTK_COLOR_SELECTION(palette_selection), c);

  return SUCCEEDED;
}


static gint palette_hide_window(GtkWidget *widget, gpointer data) {

  gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(memory_item_factory, "/Windows/Palette"), FALSE);

  return TRUE;
}
