
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <gtk/gtk.h>

#include "defines.h"
#include "memory.h"
#include "editor.h"
#include "tiled.h"
#include "palette.h"
#include "common.h"


char prefs_file_name[] = "tiledriller.conf";

/* startup attributes */
int memory_window_dx = 256, memory_window_dy = 256;
int memory_window_x = 40, memory_window_y = 40;
int editor_window_x = 70, editor_window_y = 70;
int tiled_window_x = 100, tiled_window_y = 100;
int palette_window_x = 140, palette_window_y = 140;

/* prefs status */
int _prefs_saved = NO;


static void _prefs_get_file_name(char *c) {

#ifndef WIN32
  char *e;


  e = getenv("HOME");
  if (e == NULL)
    strcpy(c, prefs_file_name);
  else {
    strcpy(c, e);
    strcat(c, "/.");
    strcat(c, prefs_file_name);
  }
#else
  strcpy(c, prefs_file_name);
#endif
}


int prefs_read(void) {

  char tmp[256];
  FILE *f;
  int i, a, b;


  /* init the file paths */
  editor_file_path[0] = 0;
  memory_file_path[0] = 0;

  _prefs_get_file_name(tmp);

  f = fopen(tmp, "rb");
  if (f == NULL)
    return FAILED;

  while (fscanf(f, "%255s", tmp) != EOF) {
    if (strcmp("Version", tmp) == 0) {
      fscanf(f, "%d", &i);
      if (i != 1) {
	fprintf(stderr, "PREFS_LOAD: Configuration file version %d is not supported.\n", i);
	fclose(f);
	return FAILED;
      }
    }
    else if (strcmp("TiledViewWindow", tmp) == 0) {
      fscanf(f, "%d", &tiled_view_status);
    }
    else if (strcmp("PaletteWindow", tmp) == 0) {
      fscanf(f, "%d", &palette_status);
    }
    else if (strcmp("MemorySnapToGrid", tmp) == 0) {
      fscanf(f, "%d", &memory_snap_to_grid);
    }
    else if (strcmp("MemoryUpdatedInRealTime", tmp) == 0) {
      fscanf(f, "%d", &memory_real_time);
    }
    else if (strcmp("TiledViewUpdatedInRealTime", tmp) == 0) {
      fscanf(f, "%d", &tiled_real_time);
    }
    else if (strcmp("TiledViewZoom", tmp) == 0) {
      fscanf(f, "%d", &tiled_zoom);
    }
    else if (strcmp("MemoryZoom", tmp) == 0) {
      fscanf(f, "%d", &memory_zoom);
    }
    else if (strcmp("EditorZoom", tmp) == 0) {
      fscanf(f, "%d", &editor_zoom);
    }
    else if (strcmp("EditorToolMode", tmp) == 0) {
      fscanf(f, "%d", &editor_tool_mode);
    }
    else if (strcmp("TiledViewSize", tmp) == 0) {
      fscanf(f, "%d %d", &tiled_tiles_dx, &tiled_tiles_dy);
    }
    else if (strcmp("MemorySize", tmp) == 0) {
      fscanf(f, "%d %d", &memory_dx, &memory_dy);
    }
    else if (strcmp("EditorSize", tmp) == 0) {
      fscanf(f, "%d %d", &editor_dx, &editor_dy);
    }
    else if (strcmp("MemoryWindowSize", tmp) == 0) {
      fscanf(f, "%d %d", &a, &b);
      if (a >= 0 && b >= 0) {
	memory_window_dx = a;
	memory_window_dy = b;
      }
    }
    else if (strcmp("TiledViewViewMode", tmp) == 0) {
      fscanf(f, "%d", &tiled_view_mode);
    }
    else if (strcmp("MemoryViewMode", tmp) == 0) {
      fscanf(f, "%d", &memory_view_mode);
    }
    else if (strcmp("EditorViewMode", tmp) == 0) {
      fscanf(f, "%d", &editor_view_mode);
    }
    else if (strcmp("MemoryWindowOrigin", tmp) == 0) {
      fscanf(f, "%d %d", &a, &b);
      if (a >= 0 && b >= 0) {
	memory_window_x = a;
	memory_window_y = b;
      }
    }
    else if (strcmp("EditorWindowOrigin", tmp) == 0) {
      fscanf(f, "%d %d", &a, &b);
      if (a >= 0 && b >= 0) {
	editor_window_x = a;
	editor_window_y = b;
      }
    }
    else if (strcmp("TiledViewWindowOrigin", tmp) == 0) {
      fscanf(f, "%d %d", &a, &b);
      if (a >= 0 && b >= 0) {
	tiled_window_x = a;
	tiled_window_y = b;
      }
    }
    else if (strcmp("PaletteWindowOrigin", tmp) == 0) {
      fscanf(f, "%d %d", &a, &b);
      if (a >= 0 && b >= 0) {
	palette_window_x = a;
	palette_window_y = b;
      }
    }
    else if (strcmp("PaletteWindowBGColor", tmp) == 0) {
      fscanf(f, "%d", &i);
      editor_pen_bg_r = i;
      fscanf(f, "%d", &i);
      editor_pen_bg_g = i;
      fscanf(f, "%d", &i);
      editor_pen_bg_b = i;
      fscanf(f, "%d", &i);
      editor_pen_bg_a = i;
    }
    else if (strcmp("PaletteWindowFGColor", tmp) == 0) {
      fscanf(f, "%d", &i);
      editor_pen_r = i;
      fscanf(f, "%d", &i);
      editor_pen_g = i;
      fscanf(f, "%d", &i);
      editor_pen_b = i;
      fscanf(f, "%d", &i);
      editor_pen_a = i;
    }
    else if (strcmp("EditorFilePath", tmp) == 0) {
      while (fscanf(f, "%c", tmp) != EOF) {
	if (tmp[0] == '"')
	  break;
      }
      for (i = 0; i < 255; i++) {
	if (fscanf(f, "%c", tmp) == EOF)
	  break;
	if (tmp[0] == '"')
	  break;
	editor_file_path[i] = tmp[0];
      }
      editor_file_path[i] = 0;
    }
    else if (strcmp("MemoryFilePath", tmp) == 0) {
      while (fscanf(f, "%c", tmp) != EOF) {
	if (tmp[0] == '"')
	  break;
      }
      for (i = 0; i < 255; i++) {
	if (fscanf(f, "%c", tmp) == EOF)
	  break;
	if (tmp[0] == '"')
	  break;
	memory_file_path[i] = tmp[0];
      }
      memory_file_path[i] = 0;
    }
  }

  fclose(f);

  return SUCCEEDED;
}


int prefs_save(void) {

  char tmp[256];
  FILE *f;


  if (_prefs_saved == YES)
    return FAILED;

  _prefs_get_file_name(tmp);

  f = fopen(tmp, "wb");
  if (f == NULL) {
    fprintf(stderr, "PREFS_SAVE: Could not open file \"%s\" for writing.\n", tmp);
    return FAILED;
  }

  gdk_window_get_size(memory_window->window, &memory_window_dx, &memory_window_dy);
  gtk_window_get_position(GTK_WINDOW(memory_window), &memory_window_x, &memory_window_y);
  gtk_window_get_position(GTK_WINDOW(editor_window), &editor_window_x, &editor_window_y);
  if (tiled_view_status == ON)
    gtk_window_get_position(GTK_WINDOW(tiled_window), &tiled_window_x, &tiled_window_y);
  if (palette_status == ON)
    gtk_window_get_position(GTK_WINDOW(palette_window), &palette_window_x, &palette_window_y);

  fprintf(f, "\n;\n; Tile Driller Configuration File\n;\n\n");
  fprintf(f, "Version 1\n");
  fprintf(f, "\n");
  fprintf(f, "MemoryFilePath \"%s\"\n", memory_file_path);
  fprintf(f, "MemorySize %d %d\n", memory_dx, memory_dy);
  fprintf(f, "MemorySnapToGrid %d\n", memory_snap_to_grid);
  fprintf(f, "MemoryUpdatedInRealTime %d\n", memory_real_time);
  fprintf(f, "MemoryViewMode %d\n", memory_view_mode);
  fprintf(f, "MemoryWindowOrigin %d %d\n", memory_window_x, memory_window_y);
  fprintf(f, "MemoryWindowSize %d %d\n", memory_window_dx, memory_window_dy);
  fprintf(f, "MemoryZoom %d\n", memory_zoom);
  fprintf(f, "\n");
  fprintf(f, "EditorFilePath \"%s\"\n", editor_file_path);
  fprintf(f, "EditorSize %d %d\n", editor_dx, editor_dy);
  fprintf(f, "EditorToolMode %d\n", editor_tool_mode);
  fprintf(f, "EditorViewMode %d\n", editor_view_mode);
  fprintf(f, "EditorWindowOrigin %d %d\n", editor_window_x, editor_window_y);
  fprintf(f, "EditorZoom %d\n", editor_zoom);
  fprintf(f, "\n");
  fprintf(f, "TiledViewSize %d %d\n", tiled_tiles_dx, tiled_tiles_dy);
  fprintf(f, "TiledViewUpdatedInRealTime %d\n", tiled_real_time);
  fprintf(f, "TiledViewViewMode %d\n", tiled_view_mode);
  fprintf(f, "TiledViewWindow %d\n", tiled_view_status);
  fprintf(f, "TiledViewWindowOrigin %d %d\n", tiled_window_x, tiled_window_y);
  fprintf(f, "TiledViewZoom %d\n", tiled_zoom);
  fprintf(f, "\n");
  fprintf(f, "PaletteWindow %d\n", palette_status);
  fprintf(f, "PaletteWindowBGColor %d %d %d %d\n", editor_pen_bg_r, editor_pen_bg_g, editor_pen_bg_b, editor_pen_bg_a);
  fprintf(f, "PaletteWindowFGColor %d %d %d %d\n", editor_pen_r, editor_pen_g, editor_pen_b, editor_pen_a);
  fprintf(f, "PaletteWindowOrigin %d %d\n", palette_window_x, palette_window_y);
  fprintf(f, "\n");

  fclose(f);

  /* save only once */
  _prefs_saved = YES;

  return SUCCEEDED;
}


int prefs_apply(void) {

  /* init the memory window data structures */
  memory_resize_data(memory_dx, memory_dy);
  memory_resize_view(memory_zoom);
  common_copy_data_to_view(memory_data, memory_view, memory_dx, memory_dy, memory_zoom, memory_view_mode);

  /* init the data structures */
  editor_resize_data(editor_dx, editor_dy);
  editor_resize_view(editor_zoom);
  memory_copy_data_out(editor_data, memory_x, memory_y, editor_dx, editor_dy);
  common_copy_data_to_view(editor_data, editor_view, editor_dx, editor_dy, editor_zoom, editor_view_mode);

  /* init the tiled view data structures */
  tiled_resize_data(tiled_dx, tiled_dy);
  tiled_resize_view(tiled_zoom);
  tiled_window_import_data();
  common_copy_data_to_view(tiled_data, tiled_view, tiled_dx, tiled_dy, tiled_zoom, tiled_view_mode);

  /* tiled updated-in-real-time */
  if (tiled_real_time == YES) {
    tiled_real_time = NO;
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(tiled_item_factory, "/View/Updated in Real-Time"), TRUE);
  }

  /* memory updated-in-real-time */
  if (memory_real_time == YES) {
    memory_real_time = NO;
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(memory_item_factory, "/View/Updated in Real-Time"), TRUE);
  }

  /* memory snap-to-grid */
  if (memory_snap_to_grid == YES) {
    memory_snap_to_grid = NO;
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(memory_item_factory, "/View/Snap to Grid"), TRUE);
  }

  /* memory window attributes */
  gtk_window_move(GTK_WINDOW(memory_window), memory_window_x, memory_window_y);
  gtk_window_resize(GTK_WINDOW(memory_window), memory_window_dx, memory_window_dy);

  /* editor window attributes */
  gtk_window_move(GTK_WINDOW(editor_window), editor_window_x, editor_window_y);

  /* palette window attributes */
  gtk_window_move(GTK_WINDOW(palette_window), palette_window_x, palette_window_y);

  /* tiled view window attributes */
  gtk_window_move(GTK_WINDOW(tiled_window), tiled_window_x, tiled_window_y);

  /* memory file paths */
  if (memory_file_path[0] != 0) {
    gtk_file_selection_set_filename(GTK_FILE_SELECTION(memory_file_selection_save), memory_file_path);
    gtk_file_selection_set_filename(GTK_FILE_SELECTION(memory_file_selection_open), memory_file_path);
  }

  /* editor file paths */
  if (editor_file_path[0] != 0) {
    gtk_file_selection_set_filename(GTK_FILE_SELECTION(editor_file_selection_save), editor_file_path);
    gtk_file_selection_set_filename(GTK_FILE_SELECTION(editor_file_selection_open), editor_file_path);
  }

  /* editor tool mode */
  if (editor_tool_mode == TOOL_MODE_BLUR_GAUSSIAN)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(editor_item_factory, "/Tools/Blur/Gaussian"), TRUE);
  else if (editor_tool_mode == TOOL_MODE_BLUR_NORMAL)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(editor_item_factory, "/Tools/Blur/Normal"), TRUE);
  else if (editor_tool_mode == TOOL_MODE_FILL)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(editor_item_factory, "/Tools/Fill"), TRUE);
  else if (editor_tool_mode == TOOL_MODE_PEN)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(editor_item_factory, "/Tools/Pen"), TRUE);
  else if (editor_tool_mode == TOOL_MODE_PICK)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(editor_item_factory, "/Tools/Pick"), TRUE);

  /* editor size */
  if (editor_dx == 8)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(editor_item_factory, "/Size/8x8"), TRUE);
  else if (editor_dx == 16)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(editor_item_factory, "/Size/16x16"), TRUE);
  else if (editor_dx == 32)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(editor_item_factory, "/Size/32x32"), TRUE);
  else if (editor_dx == 64)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(editor_item_factory, "/Size/64x64"), TRUE);
  else if (editor_dx == 128)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(editor_item_factory, "/Size/128x128"), TRUE);
  else if (editor_dx == 256)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(editor_item_factory, "/Size/256x256"), TRUE);
  else if (editor_dx == 512)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(editor_item_factory, "/Size/512x512"), TRUE);
  else if (editor_dx == 1024)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(editor_item_factory, "/Size/1024x1024"), TRUE);

  /* editor zoom */
  if (editor_zoom == 1)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(editor_item_factory, "/Zoom/100%"), TRUE);
  else if (editor_zoom == 2)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(editor_item_factory, "/Zoom/200%"), TRUE);
  else if (editor_zoom == 4)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(editor_item_factory, "/Zoom/400%"), TRUE);
  else if (editor_zoom == 6)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(editor_item_factory, "/Zoom/600%"), TRUE);
  else if (editor_zoom == 10)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(editor_item_factory, "/Zoom/1000%"), TRUE);
  else if (editor_zoom == 15)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(editor_item_factory, "/Zoom/1500%"), TRUE);
  else if (editor_zoom == 20)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(editor_item_factory, "/Zoom/2000%"), TRUE);

  /* tiled view size */
  if (tiled_tiles_dx == 3)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(tiled_item_factory, "/Size/3x3"), TRUE);
  else if (tiled_tiles_dx == 5)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(tiled_item_factory, "/Size/5x5"), TRUE);
  else if (tiled_tiles_dx == 7)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(tiled_item_factory, "/Size/7x7"), TRUE);
  else if (tiled_tiles_dx == 9)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(tiled_item_factory, "/Size/9x9"), TRUE);

  /* tiled view zoom */
  if (tiled_zoom == 1)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(tiled_item_factory, "/Zoom/100%"), TRUE);
  else if (tiled_zoom == 2)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(tiled_item_factory, "/Zoom/200%"), TRUE);
  else if (tiled_zoom == 3)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(tiled_item_factory, "/Zoom/300%"), TRUE);
  else if (tiled_zoom == 4)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(tiled_item_factory, "/Zoom/400%"), TRUE);

  /* memory size */
  if (memory_dx == 64)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(memory_item_factory, "/Size/64x64"), TRUE);
  else if (memory_dx == 128)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(memory_item_factory, "/Size/128x128"), TRUE);
  else if (memory_dx == 256)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(memory_item_factory, "/Size/256x256"), TRUE);
  else if (memory_dx == 512)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(memory_item_factory, "/Size/512x512"), TRUE);
  else if (memory_dx == 1024)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(memory_item_factory, "/Size/1024x1024"), TRUE);
  else if (memory_dx == 2048)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(memory_item_factory, "/Size/2048x2048"), TRUE);

  /* memory zoom */
  if (memory_zoom == 1)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(memory_item_factory, "/Zoom/100%"), TRUE);
  else if (memory_zoom == 2)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(memory_item_factory, "/Zoom/200%"), TRUE);
  else if (memory_zoom == 3)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(memory_item_factory, "/Zoom/300%"), TRUE);
  else if (memory_zoom == 4)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(memory_item_factory, "/Zoom/400%"), TRUE);

  /* memory view mode */
  if (memory_view_mode == VIEW_MODE_RGBA)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(memory_item_factory, "/View/RGBA"), TRUE);
  else if (memory_view_mode == VIEW_MODE_RGB)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(memory_item_factory, "/View/RGB"), TRUE);
  else if (memory_view_mode == VIEW_MODE_ALPHA)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(memory_item_factory, "/View/Alpha"), TRUE);

  /* editor view mode */
  if (editor_view_mode == VIEW_MODE_RGBA)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(editor_item_factory, "/View/RGBA"), TRUE);
  else if (editor_view_mode == VIEW_MODE_RGB)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(editor_item_factory, "/View/RGB"), TRUE);
  else if (editor_view_mode == VIEW_MODE_ALPHA)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(editor_item_factory, "/View/Alpha"), TRUE);

  /* tiled view view mode */
  if (tiled_view_mode == VIEW_MODE_RGBA)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(tiled_item_factory, "/View/RGBA"), TRUE);
  else if (tiled_view_mode == VIEW_MODE_RGB)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(tiled_item_factory, "/View/RGB"), TRUE);
  else if (tiled_view_mode == VIEW_MODE_ALPHA)
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(tiled_item_factory, "/View/Alpha"), TRUE);

  /* display the windows */
  gtk_widget_show_all(memory_window);
  gtk_widget_show_all(editor_window);

  /* palette window */
  if (palette_status == ON) {
    palette_status = OFF;
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(memory_item_factory, "/Windows/Palette"), TRUE);
  }

  /* tiled view window */
  if (tiled_view_status == ON) {
    tiled_view_status = OFF;
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(memory_item_factory, "/Windows/Tiled View"), TRUE);
  }

  /* update the color selector */
  palette_window_refresh();

  return SUCCEEDED;
}
